from __future__ import annotations

from .abstracts import AbstractInterface

PACKAGE_NAME = __name__
BACKEND_FOLDER_NAME = "backends"
USER_BACKENDS_FOLDER_NAME = "user_backends"
CUSTOM_BACKEND_FOLDER_NAME = "custom_backends"
TEMPLATES_FOLDER_NAME = "templates"

BASE_BACKEND_MODULE = f"{PACKAGE_NAME}.{BACKEND_FOLDER_NAME}"
USER_BACKEND_MODULE = f"{PACKAGE_NAME}.{USER_BACKENDS_FOLDER_NAME}"



from __future__ import annotations

from logging import getLogger
from typing import Any, Callable

import torch
from qadence2_ir.types import Assign, Call, Load, Model

logger = getLogger(__name__)


def torch_call(call: Call) -> Callable[[dict, dict], torch.Tensor]:
    """
    Convert a `Call` object into a torchified function which can be evaluated using.

    a vparams and inputs dict.
    """
    fn = getattr(torch, call.identifier)

    def evaluate(params: dict, inputs: dict) -> torch.Tensor:
        args = []
        for symbol in call.args:
            if isinstance(symbol, float):
                # NOTE we compile constants into each TorchCallable instead of passing
                # them around in the values dict
                args.append(torch.tensor(symbol))
            elif isinstance(symbol, Load):
                args.append({**params, **inputs}[symbol.variable])
        return fn(*args)

    return evaluate


class ParameterBuffer(torch.nn.Module):
    """
    A class holding all root parameters either passed by the user.

    or trainable variational parameters.
    """

    def __init__(
        self,
        trainable_vars: list[str],
        non_trainable_vars: list[str],
    ) -> None:
        super().__init__()
        self.vparams = {p: torch.rand(1, requires_grad=True) for p in trainable_vars}
        self.fparams = {p: None for p in non_trainable_vars}
        self._dtype = torch.float64
        self._device = torch.device("cpu")

    @property
    def device(self) -> torch.device:
        return self._device

    @property
    def dtype(self) -> torch.dtype:
        return self._dtype

    def to(self, args: Any, kwargs: Any) -> None:
        self.vparams = {p: t.to(*args, **kwargs) for p, t in self.vparams.items()}
        try:
            k = next(iter(self.vparams))
            t = self.vparams[k]
            self._device = t.device
            self._dtype = t.dtype
        except Exception:
            pass

    @classmethod
    def from_model(cls, model: Model) -> ParameterBuffer:
        f_p = []
        v_p = []
        for param_name, alloc in model.inputs.items():
            if alloc.is_trainable:
                v_p.append(param_name)
            else:
                f_p.append(param_name)
        return ParameterBuffer(v_p, f_p)


class Embedding(torch.nn.Module):
    """
    A class holding:

    - A parameterbuffer containing concretized vparams + list of featureparams,
    - A dictionary of intermediate and leaf variable names mapped to a TorchCall object
        which can be results of function/expression evaluations.
    """

    def __init__(self, model: Model) -> None:
        super().__init__()
        self.param_buffer = ParameterBuffer.from_model(model)
        self.var_to_torchcall: dict[str, Callable] = self.create_var_to_torchcall_mapping(model)

    def __call__(self, inputs: dict[str, torch.Tensor]) -> dict[str, torch.Tensor]:
        """
        Expects a dict of user-passed name:value pairs for featureparameters.

        and assigns all intermediate and leaf variables using the current vparam values
        and the passed values for featureparameters.
        """
        assigned_params: dict[str, torch.Tensor] = {}
        try:
            assert inputs.keys() == self.param_buffer.fparams.keys()
        except Exception as _:
            logger.error("Please pass a dict containing name:value for each fparam.")
        for var, torchcall in self.var_to_torchcall.items():
            assigned_params[var] = torchcall(
                self.param_buffer.vparams,
                {
                    **inputs,
                    **assigned_params,
                },  # we add the "intermediate" variables too
            )
        assigned_params = {**assigned_params, **inputs}
        return assigned_params

    @staticmethod
    def create_var_to_torchcall_mapping(
        model: Model,
    ) -> dict[str, Callable[[dict, dict], torch.Tensor]]:
        assign_to_torch = dict()
        for instr in model.instructions:
            if isinstance(instr, Assign):
                assign_to_torch[instr.variable] = torch_call(instr.value)
        return assign_to_torch



from __future__ import annotations

import warnings

import numpy as np
from pulser.devices._devices import AnalogDevice
from pulser.register.register_layout import RegisterLayout
from qadence2_ir.types import Model

warnings.filterwarnings("ignore", category=UserWarning)
warnings.formatwarning = lambda msg, *args, **kwargs: f"WARNING: {msg}\n"


def from_model(model: Model) -> RegisterLayout:
    model_register = model.register

    if model_register.num_qubits < 1:
        raise ValueError("No qubit available in the register.")

    if model_register.grid_type and model_register.grid_type != "triangular":
        warnings.warn(
            "Fresnel-1 only supoorts triangular grids at the moment.",
            SyntaxWarning,
            stacklevel=2,
        )

    if model_register.grid_scale != 1.0:
        warnings.warn(
            "Currently, Fresnel-1 uses a fixed grid.",
            SyntaxWarning,
            stacklevel=2,
        )

    if model.directives.get("enable_digital_analog"):
        warnings.warn(
            "Fresnel-1 uses a fixed grid and does not have digital channels.\n"
            + "Digital operations using atomic distance-based strategies\n"
            + "may not behave as expected.",
            SyntaxWarning,
            stacklevel=2,
        )

    coords = model_register.qubit_positions

    if not coords:
        shift = model_register.num_qubits // 2
        coords = [(p - shift, 0) for p in range(model_register.num_qubits)]

    transform = np.array([[1.0, 0.0], [0.5, 0.8660254037844386]])
    coords = coords @ transform

    layout = AnalogDevice.calibrated_register_layouts["TriangularLatticeLayout(61, 5.0µm)"]
    coords *= 5
    traps = layout.get_traps_from_coordinates(*coords)
    register = layout.define_register(*traps, qubit_ids=range(len(traps)))

    return register  # type: ignore



from __future__ import annotations

from collections import Counter
from typing import Any, Callable, Optional, Union, cast

from pulser.sequence.sequence import Sequence
from pulser_simulation.simresults import SimulationResults
from pulser_simulation.simulation import QutipEmulator
from qutip import Qobj

from qadence2_platforms import AbstractInterface
from qadence2_platforms.abstracts import OnEnum, RunEnum

RunResult = Union[Counter, Qobj]


class Interface(AbstractInterface[float, Sequence, float, RunResult, Counter, Qobj]):

    def __init__(self, sequence: Sequence, non_trainable_parameters: set[str]) -> None:
        self._non_trainable_parameters = non_trainable_parameters
        self._params: dict[str, float] = dict()
        self._sequence = sequence

    @property
    def info(self) -> dict[str, Any]:
        return {"device": self.sequence.device, "register": self.sequence.register}

    @property
    def sequence(self) -> Sequence:
        return self._sequence

    def set_parameters(self, params: dict[str, float]) -> None:
        valid_params = params.keys() & self._non_trainable_parameters

        if valid_params != params.keys():
            raise ValueError(f"{set(params.keys())} are not fixed parameters in this sequence.")

        self._params = params

    def _run(
        self,
        run_type: RunEnum,
        platform: SimulationResults,
        shots: int | None = None,
        observable: Any | None = None,
        callback: Optional[Callable] = None,
        **_: Any,
    ) -> Any:
        """
        Method to be used for `_on_emulator` or `_on_qpu` to execute the respective.

        `run_type` option.

        **Notice**: for now, it only supports `emulator` option and
        `QutipEmulator` platform.

        :param run_type: str: `run`, `sample`, `expectation` options
        :param platform: callable to retrieve methods for executing the options above
        :param shots: number of shots, if applicable (`sample` only)
        :param observable: list of observables, if applicable (`expectation` only)
        :param callback: callback function to be used inside the method, if applicable
        :return: the respective result value: `Qobj` for `run`, `Counter` for `sample`,
            and numeric type (`float`, `complex`, `ArrayLike`) for `expectation`
        """

        match run_type:
            case RunEnum.RUN:
                return platform.get_final_state()
            case RunEnum.SAMPLE:
                return platform.sample_final_state(shots)
            case RunEnum.EXPECTATION:
                return platform.expect(obs_list=observable)
            case _:
                raise NotImplementedError(f"Run type '{run_type}' not implemented.")

    def _on_emulator(
        self,
        run_type: RunEnum,
        values: dict[str, float] | None,
        shots: int | None = None,
        observable: Any | None = None,
        callback: Optional[Callable] = None,
        **_: Any,
    ) -> Any:
        """
        Runs on Pulser emulator.

        **Notice**: for now, only `QutipEmulator` is supported.

        It uses `run_type` to decide which kind of run it will perform: `run`, `sample`,
        or `expectation`.

        :param run_type: str: `run`, `sample`, `expectation` possible values
        :param values: dictionary of user-input parameters
        :param shots: int: number of shots; applied only for `sample` option
        :param observable: list of observables; applied only for `expectation` option
        :param callback: callback function to be used inside the method (if applicable)
        :return: the respective result value: `Qobj` for `run`, `Counter` for `sample`,
            and numeric type (`float`, `complex`, `ArrayLike`) for `expectation`
        """
        vals: dict[str, float] = {**(values or dict()), **self._params}
        pulse_sequence: Sequence = self.sequence.build(**vals)  # type: ignore
        simulation: QutipEmulator = QutipEmulator.from_sequence(
            pulse_sequence, with_modulation=True
        )
        result: SimulationResults = simulation.run()

        return self._run(
            run_type=run_type,
            platform=result,
            shots=shots,
            observable=observable,
            callback=callback,
        )

    def _on_qpu(
        self,
        run_type: RunEnum,
        values: dict[str, float] | None,
        shots: int | None = None,
        observable: Any | None = None,
        callback: Optional[Callable] = None,
        **_: Any,
    ) -> Any:
        """
        Runs on available QPU.

        **Notice**: for now, it is not implemented yet.

        :param run_type: str: `run`, `sample`, `expectation` possible values
        :param values: dictionary of user-input parameters
        :param shots: number of shots
        :param observable: list of observables; applied only for `expectation` option
        :param callback: callback function to be used inside the method (if applicable)
        :return: the respective result value: `Qobj` for `run`, `Counter` for `sample`,
            and numeric type (`float`, `complex`, `ArrayLike`) for `expectation`
        """
        raise NotImplementedError("QPU execution not implemented yet.")

    def run(
        self,
        *,
        values: dict[str, float] | None = None,
        on: OnEnum = OnEnum.EMULATOR,
        shots: int | None = None,
        callback: Optional[Callable] = None,
        **_: Any,
    ) -> RunResult:
        match on:
            case OnEnum.EMULATOR:
                return self._on_emulator(
                    run_type=RunEnum.RUN, values=values, shots=shots, callback=callback
                )
            case OnEnum.QPU:
                return self._on_qpu(
                    run_type=RunEnum.RUN, values=values, shots=shots, callback=callback
                )
            case _:
                raise NotImplementedError(f"Platform '{on}' not implemented.")

    def sample(
        self,
        *,
        values: dict[str, float] | None = None,
        shots: int | None = None,
        on: OnEnum = OnEnum.EMULATOR,
        callback: Optional[Callable] = None,
        **_: Any,
    ) -> Counter:
        match on:
            case OnEnum.EMULATOR:
                return cast(
                    Counter,
                    self._on_emulator(
                        run_type=RunEnum.SAMPLE,
                        values=values,
                        shots=shots,
                        callback=callback,
                    ),
                )
            case OnEnum.QPU:
                return cast(
                    Counter,
                    self._on_qpu(
                        run_type=RunEnum.SAMPLE,
                        values=values,
                        shots=shots,
                        callback=callback,
                    ),
                )
            case _:
                raise NotImplementedError(f"Platform '{on}' not implemented.")

    def expectation(
        self,
        *,
        values: dict[str, float] | None = None,
        on: OnEnum = OnEnum.EMULATOR,
        shots: int | None = None,
        observable: Any | None = None,
        callback: Optional[Callable] = None,
        **_: Any,
    ) -> Qobj:
        match on:
            case OnEnum.EMULATOR:
                return self._on_emulator(
                    run_type=RunEnum.EXPECTATION,
                    values=values,
                    shots=shots,
                    observable=observable,
                    callback=callback,
                )
            case OnEnum.QPU:
                return self._on_qpu(
                    run_type=RunEnum.EXPECTATION,
                    values=values,
                    shots=shots,
                    observable=observable,
                    callback=callback,
                )
            case _:
                raise NotImplementedError(f"Platform '{on}' not implemented.")



from __future__ import annotations

from logging import getLogger
from typing import Any, Callable, Counter, Literal, Optional, cast

import pyqtorch as pyq
import torch

from qadence2_platforms.abstracts import (
    AbstractInterface,
    RunEnum,
)

from .embedding import Embedding
from .functions import load_observables
from .register import RegisterInterface
from .utils import InputType

logger = getLogger(__name__)


class Interface(
    AbstractInterface[
        torch.Tensor,
        pyq.QuantumCircuit,
        torch.Tensor,
        torch.Tensor,
        list[Counter],
        torch.Tensor,
    ],
):
    """A class holding register, embedding, circuit, native backends and optional observable."""

    def __init__(
        self,
        register: RegisterInterface,
        embedding: Embedding,
        circuit: pyq.QuantumCircuit,
        observable: list[InputType] | InputType | None = None,
    ) -> None:
        super().__init__()
        self.register = register
        self.init_state: torch.Tensor = (
            circuit.from_bitstring(register.init_state)
            if register.init_state is not None
            else circuit.init_state()
        )
        self.embedding = embedding
        self.circuit = circuit
        self.observable = observable

    @property
    def info(self) -> dict[str, Any]:
        return {"num_qubits": self.register.n_qubits}

    @property
    def sequence(self) -> pyq.QuantumCircuit:
        return self.circuit

    def add_noise(self, model: Literal["SPAM"]) -> None:
        pass

    def set_parameters(self, params: dict[str, float]) -> None:
        pass

    def _run(
        self,
        run_type: RunEnum,
        values: dict[str, torch.Tensor] | None = None,
        callback: Optional[Callable] = None,
        state: torch.Tensor | None = None,
        shots: int | None = None,
        observable: list[InputType] | InputType | None = None,
        **_: Any,
    ) -> Any:
        """
        Method to execute run type-specific option.

        Option can be `run`, `sample`
        or `expectation`. Each option can have different arguments, such as:
        `sample` uses `shots`, while `expectation` uses `observable`.

        It should not be called directly. Use it on `run`, `sample` or `expectation`
        methods.

        :param run_type: str option as `run`, `sample` or `expectation`
        :param values: dictionary of user-input parameters
        :param callback: callback function to be used internally, if applicable
        :param state: a tensor containing the desired state to perform the execution from
        :param shots: number of shots, if applicable (`sample` only)
        :param observable: a list of observables, if applicable (`expectation` only)
        :return: a tensor or list of values (`sample` only) of the calculated state
        """
        inputs = values or dict()
        state = state if state is not None else self.init_state

        match run_type:
            case RunEnum.RUN:
                return pyq.run(
                    circuit=self.circuit,
                    state=state,
                    values=inputs,
                    embedding=self.embedding,
                )
            case RunEnum.SAMPLE:
                return pyq.sample(
                    circuit=self.circuit,
                    state=state,
                    values=inputs,
                    n_shots=shots,
                    embedding=self.embedding,
                )
            case RunEnum.EXPECTATION:
                if observable is not None or self.observable is not None:
                    return pyq.expectation(
                        circuit=self.circuit,
                        state=state,
                        values=inputs,
                        observable=load_observables(observable or self.observable),  # type: ignore [arg-type]
                        embedding=self.embedding,
                    )
                raise ValueError("Observable must not be None for expectation run.")
            case _:
                raise NotImplementedError(f"Run type '{run_type}' not implemented.")

    def run(
        self,
        *,
        values: dict[str, torch.Tensor] | None = None,
        callback: Optional[Callable] = None,
        state: torch.Tensor | None = None,
        **kwargs: Any,
    ) -> torch.Tensor:
        return self._run(RunEnum.RUN, values=values, callback=callback, state=state, **kwargs)

    def sample(
        self,
        *,
        values: dict[str, torch.Tensor] | None = None,
        shots: int | None = None,
        callback: Optional[Callable] = None,
        state: torch.Tensor | None = None,
        **kwargs: Any,
    ) -> list[Counter]:
        return cast(
            list,
            self._run(
                RunEnum.SAMPLE,
                values=values,
                callback=callback,
                shots=shots,
                state=state,
                **kwargs,
            ),
        )

    def expectation(
        self,
        *,
        values: dict[str, torch.Tensor] | None = None,
        callback: Optional[Callable] = None,
        state: torch.Tensor | None = None,
        observable: list[InputType] | InputType | None = None,
        **kwargs: Any,
    ) -> torch.Tensor:
        return self._run(
            RunEnum.EXPECTATION,
            values=values,
            callback=callback,
            state=state,
            observable=observable,
            **kwargs,
        )



from __future__ import annotations

from dataclasses import dataclass
from logging import getLogger

import pyqtorch as pyq
from qadence2_ir.types import Load, Model, QuInstruct

from qadence2_platforms.backends.pyqtorch.embedding import Embedding
from qadence2_platforms.backends.pyqtorch.interface import Interface
from qadence2_platforms.backends.pyqtorch.register import RegisterInterface

logger = getLogger(__name__)


@dataclass(frozen=True)
class Compiler:
    instruction_mapping = {
        "not": pyq.CNOT,
        "add": pyq.Add,
        "mul": pyq.Scale,
        "noncommute": pyq.Sequence,
    }

    def compile(
        self,
        model: Model,
    ) -> pyq.QuantumCircuit:
        pyq_operations = []
        for instr in model.instructions:
            if isinstance(instr, QuInstruct):
                native_op = None
                try:
                    native_op = getattr(pyq, instr.name.upper())
                except Exception as _:
                    native_op = self.instruction_mapping[instr.name]
                control = instr.support.control
                target = instr.support.target
                native_support = (*control, *target)
                if len(instr.args) > 0:
                    assert len(instr.args) == 1, "More than one arg not supported"
                    (maybe_load,) = instr.args
                    assert isinstance(maybe_load, Load), "only support load"
                    pyq_operations.append(native_op(native_support, maybe_load.variable))
                else:
                    pyq_operations.append(native_op(*native_support))
        return pyq.QuantumCircuit(model.register.num_qubits, pyq_operations)


def compile_to_backend(model: Model) -> Interface:
    register_interface = RegisterInterface(
        model.register.num_qubits, model.register.options.get("init_state")
    )
    embedding = Embedding(model)
    native_circ = Compiler().compile(model)
    return Interface(register_interface, embedding, native_circ)



# type: ignore
# TODO: 1. remove the line above after implementing the backend
# TODO: 2. remove these comments

from __future__ import annotations

import logging
from typing import Any, Callable, Optional, TypeVar

from qadence2_platforms.abstracts import (
    AbstractInterface,
)

logger = logging.getLogger(__name__)


# TODO: replace the following types by relevant types for this custom interface
# TODO: erase these TODO comments
ArrayType = TypeVar("ArrayType")
SequenceType = TypeVar("SequenceType")
ParameterType = TypeVar("ParameterType")
RunResultType = TypeVar("RunResultType")
SampleResultType = TypeVar("SampleResultType")
ExpectationResultType = TypeVar("ExpectationResultType")


# TODO: replace the new types in the respective order in the `AbstractInterface` below:
# TODO: erase these TODO comments
class Interface(
    AbstractInterface[
        ArrayType,
        SequenceType,
        ParameterType,
        RunResultType,
        SampleResultType,
        ExpectationResultType,
    ],
):
    """
    TODO: 1.

    fill the methods with relevant arguments and code logic

    Hint: Check the built-in backends for reference on how to write this class.

    TODO: 2. replace this docstring with a description of this custom interface class
    """

    def __init__(self) -> None:
        # TODO: 1. implement this method logic
        # TODO: 2. include relevant arguments on the method
        # TODO: 3. write docstring if needed
        # TODO: 4. erase these TODO comments
        pass

    @property
    def info(self) -> dict[str, Any]:
        # TODO: 1. implement this method logic
        # TODO: 2. erase these TODO comments
        pass

    @property
    def sequence(self) -> SequenceType:
        # TODO: 1. implement this method logic
        # TODO: 2. erase these TODO comments
        pass

    def set_parameters(self, params: dict[str, ParameterType]) -> None:
        # TODO: 1. implement this method logic
        # TODO: 2. include relevant arguments on the method
        # TODO: 3. do not remove the existing arguments!
        # TODO: 4. replace return generic type for relevant type
        # TODO: 5. write docstring if needed
        # TODO: 6. erase these TODO comments
        pass

    def run(
        self,
        *,
        values: dict[str, ArrayType] | None = None,
        callback: Optional[Callable] = None,
        **kwargs: Any,
    ) -> RunResultType:
        # TODO: 1. implement this method logic
        # TODO: 2. include relevant arguments on the method
        # TODO: 3. do not remove the existing arguments!
        # TODO: 4. replace return generic type for relevant type
        # TODO: 5. write docstring!
        # TODO: 6. erase these TODO comments
        pass

    def sample(
        self,
        *,
        values: dict[str, ArrayType] | None = None,
        shots: int | None = None,
        callback: Optional[Callable] = None,
        **kwargs: Any,
    ) -> SampleResultType:
        # TODO: 1. implement this method logic
        # TODO: 2. include relevant arguments on the method
        # TODO: 3. do not remove the existing arguments!
        # TODO: 4. replace return generic type for relevant type
        # TODO: 5. write docstring!
        # TODO: 6. erase these TODO comments
        pass

    def expectation(
        self,
        *,
        values: dict[str, ArrayType] | None = None,
        observable: Any | None = None,
        callback: Optional[Callable] = None,
        **kwargs: Any,
    ) -> ExpectationResultType:
        # TODO: 1. implement this method logic
        # TODO: 2. include relevant arguments on the method
        # TODO: 3. do not remove the existing arguments!
        # TODO: 4. replace return generic type for relevant type
        # TODO: 5. write docstring!
        # TODO: 6. erase these TODO comments
        pass



from __future__ import annotations

from typing import Any, Protocol


class Support(Protocol):
    def target_all(self) -> Any: ...

    @property
    def target(self) -> list[int]: ...

    @property
    def control(self) -> list[int]: ...


class InputType(Protocol):
    @property
    def head(self) -> InputType | str: ...

    @property
    def args(self) -> InputType | list[InputType | str | Support]: ...

    @property
    def value(self) -> Any: ...

    def is_symbol(self) -> bool: ...

    def is_quantum_operator(self) -> bool: ...

    def __getitem__(self, item: slice | int) -> Any: ...



from __future__ import annotations

from dataclasses import replace
from functools import reduce
from typing import Any, Callable, Optional

import numpy as np
from pulser.channels import DMM
from pulser.devices._devices import AnalogDevice
from pulser.parametrized.variable import VariableItem
from pulser.register.register_layout import RegisterLayout
from pulser.sequence.sequence import Sequence
from qadence2_ir.types import Alloc, Assign, Call, Load, Model, QuInstruct

from . import functions as add_pulse
from .functions import PULSE_FN_MAP


class NamedPulse:
    def __init__(self, name: str, *args: Any) -> None:
        self.name = name
        self.args = args


# Fresnel1 = AnalogDevice
Fresnel1 = replace(
    AnalogDevice.to_virtual(),
    dmm_objects=(
        DMM(
            # from Pulser tutorials/dmm.html#DMM-Channel-and-Device
            clock_period=4,
            min_duration=16,
            max_duration=2**26,
            mod_bandwidth=8,
            bottom_detuning=-2 * np.pi * 20,  # detuning between 0 and -20 MHz
            total_bottom_detuning=-2 * np.pi * 2000,  # total detuning
        ),
    ),
)


def from_model(model: Model, register: RegisterLayout) -> Sequence:
    seq = Sequence(register, Fresnel1)  # type: ignore
    seq.declare_channel("global", "rydberg_global")

    if model.directives.get("local_targets"):
        if model.directives.get("local_shifts"):
            targets = zip(
                model.directives["local_targets"],
                model.directives["local_shifts"],
            )
            detuning_map = register.define_detuning_map(
                {elem[0]: elem[1] / (2 * np.pi) for elem in targets}
            )
        else:
            targets = zip(
                model.directives["local_targets"],
                [1] * len(model.directives["local_targets"]),
            )
            detuning_map = register.define_detuning_map(dict(targets))

        seq.config_detuning_map(detuning_map, "dmm_0")

    pulses = from_instructions(seq, model.inputs, model.instructions)

    for pulse in pulses:
        fn: Optional[Callable] = getattr(
            add_pulse, PULSE_FN_MAP.get(pulse.name) or pulse.name, None
        )
        if fn is not None:
            fn(seq, *pulse.args)
        else:
            raise ValueError(f"current backend does not have pulse '{pulse.name}' implemented.")

    return seq


def from_instructions(
    sequence: Sequence,
    inputs: dict[str, Alloc],
    instructions: list[Assign | QuInstruct],
) -> list[NamedPulse]:
    variables = dict()
    temp_vars: dict[str, Any] = dict()

    for var in inputs:
        # inputs[var].size holds the points to interpolate time-dependent functions
        if inputs[var].size > 1:
            raise TypeError("This platform cannot handle time modulated variables.")

        variables[var] = sequence.declare_variable(var)

    pulses = []
    for instruction in instructions:
        if isinstance(instruction, Assign):
            assign(instruction, temp_vars, variables)

        if isinstance(instruction, QuInstruct):
            args = (
                (
                    (temp_vars.get(arg.variable) or variables[arg.variable])
                    if isinstance(arg, Load)
                    else arg
                )
                for arg in instruction.args
            )
            pulses.append(NamedPulse(instruction.name, *args))

    return pulses


def assign(
    instruction: Assign, temp_vars: dict[str, Any], variables: dict[str, VariableItem]
) -> None:
    var_name = instruction.variable

    if var_name in temp_vars or var_name in variables:
        return

    fn = instruction.value
    if isinstance(fn, Call):
        args = (
            (
                (temp_vars.get(arg.variable) or variables.get(arg.variable))
                if isinstance(arg, Load)
                else arg
            )
            for arg in fn.args
        )
        temp_vars[var_name] = _compute(fn.identifier, *args)


def _compute(fn: str, *args: Any) -> Any:
    match fn:
        case "add":
            return reduce(lambda a, b: a + b, args)
        case "mul":
            return reduce(lambda a, b: a * b, args)
        case "pow":
            return reduce(lambda a, b: a**b, args)
        case _:
            return getattr(np, fn)(*args)



from __future__ import annotations

import logging
import sys
import traceback
from importlib import import_module
from importlib.util import find_spec
from pathlib import Path
from types import ModuleType
from typing import Any

from qadence2_platforms import (
    BASE_BACKEND_MODULE,
    USER_BACKEND_MODULE,
    USER_BACKENDS_FOLDER_NAME,
)

logger = logging.getLogger(__name__)


class ModuleError(Exception):
    def __init__(self, arg: Any):
        super().__init__(arg)


def module_loader(module_name: str) -> ModuleType:
    """
    Loads an arbitrary module and returns it.

    It can be a backend submodule from
    an already imported backend, i.e. "pyqtorch", an existing but not imported
    module, i.e. "fresnel1", or a custom backend module, i.e. "custom_backend1".

    It facilitates to import and work with custom modules for backends development,
    for instance.

    :param module_name: str: The name of the module to load.
    :return: ModuleType: The loaded module.
    """

    base_backend = f"{BASE_BACKEND_MODULE}.{module_name}"
    if base_backend in sys.modules:
        return sys.modules[base_backend]
    user_backend = f"{USER_BACKEND_MODULE}.{module_name}"
    if user_backend in sys.modules:
        return sys.modules[user_backend]

    module: ModuleType
    try:
        module = import_module(base_backend)
    except ModuleNotFoundError:
        try:
            module = import_module(user_backend)
        except ModuleNotFoundError:
            traceback.print_exc()
            error_msg = (
                f"Module error. Please verify module '{module_name}'. "
                f"You may need to import it beforehand. "
                f"In case it is a custom module, please double check its path and"
                f"use `pathlib.Path` class instead of pure `str`."
            )
            logger.error(ModuleError(error_msg))
            raise ModuleError(error_msg)
    return module


def resolve_module_path(module_source: str | Path) -> bool:
    """
    Resolve module path for custom backends.

    It symlinks custom backends,
    if they are not symlinked yet, and ensure that relative imports from
    their files do not break.

    :param module_source: str or Path: module source path.
    :return: bool: True if the module path is resolved, False otherwise.
    """

    platforms_spec = find_spec("qadence2_platforms")
    if platforms_spec is None:
        return False

    if platforms_spec.origin is None:
        return False

    platforms_path = Path(platforms_spec.origin).parent
    src = Path(module_source).resolve()

    try:
        dst = platforms_path / USER_BACKENDS_FOLDER_NAME
        dst.symlink_to(src, target_is_directory=True)
    except FileExistsError:
        return True
    except Exception:
        traceback.print_exc()
        return False
    return True



from __future__ import annotations

from enum import Enum, auto
from typing import Any

import numpy as np
from pulser.parametrized.variable import VariableItem
from pulser.sequence import Sequence
from pulser.waveforms import ConstantWaveform

DEFAULT_AMPLITUDE = 4 * np.pi
DEFAULT_DETUNING = 10 * np.pi

# TODO: re-introduce `Support` to account for "local" and "global" on the `channel` arg


class Direction(Enum):
    X = auto()
    Y = auto()


class Duration(Enum):
    FILL = auto()


# pulse function mapping.
# keys are the `qadence2_ir.Model` standard, values are the function names below
PULSE_FN_MAP = {
    "rx": "rx",
    "ry": "ry",
    "not": "x",
    "h": "h",
}


def dyn_pulse(
    sequence: Sequence,
    duration: VariableItem | float,
    amplitude: VariableItem | float,
    detuning: VariableItem | float,
    phase: VariableItem | float,
    **_: Any,
) -> None:
    """
    Dynamic pulse to simulate a specific time-dependent hamiltonian for neutral-atom devices.

    :param sequence: a `pulser.sequence.Sequence` instance
    :param duration: duration of the pulse in nanoseconds
    :param amplitude: amplitude of the pulse in rad/µs
    :param detuning: detuning of the pulse in rad/s
    :param phase: phase in rad
    """
    max_amp = sequence.device.channels["rydberg_global"].max_amp or DEFAULT_AMPLITUDE
    max_abs_detuning = (
        sequence.device.channels["rydberg_global"].max_abs_detuning or DEFAULT_DETUNING
    )

    duration *= 1000 * 2 * np.pi / max_amp  # type: ignore
    amplitude *= max_amp  # type: ignore
    detuning *= max_abs_detuning  # type: ignore

    sequence.enable_eom_mode("global", amp_on=amplitude, detuning_on=detuning)
    sequence.add_eom_pulse("global", duration=duration, phase=phase)  # type: ignore
    sequence.disable_eom_mode("global")


def rx(
    sequence: Sequence,
    angle: VariableItem | float,
    **_: Any,
) -> None:
    rotation(sequence, angle, Direction.X)


def ry(
    sequence: Sequence,
    angle: VariableItem | float,
    **_: Any,
) -> None:
    rotation(sequence, angle, Direction.Y)


def x(sequence: Sequence, **_: Any) -> None:
    rotation(sequence, angle=np.pi, direction=Direction.X)


def h(
    sequence: Sequence,
    duration: VariableItem | float = 1000.0,
    **_: Any,
) -> None:
    support_list = "GLOBAL"
    amplitude = sequence.device.channels["rydberg_global"].max_amp or DEFAULT_AMPLITUDE
    duration *= 1000 * 2 * np.pi / amplitude
    duration = int(duration) if duration > 16 else 16

    sequence.enable_eom_mode("global", amp_on=amplitude, correct_phase_drift=True)
    sequence.add_eom_pulse(
        "global", duration=int(duration), phase=np.pi / 2, post_phase_shift=np.pi
    )
    sequence.disable_eom_mode(support_list)


def rotation(
    sequence: Sequence,
    angle: VariableItem | float,
    direction: Direction | float,
    **_: Any,
) -> None:

    amplitude = sequence.device.channels["rydberg_global"].max_amp or DEFAULT_AMPLITUDE
    duration = 1000 * angle / amplitude
    detuning = 0

    phase: Any
    match direction:
        case Direction.X:
            phase = 0
        case Direction.Y:
            phase = np.pi / 2
        case _:
            phase = direction

    sequence.enable_eom_mode("global", amp_on=amplitude, detuning_on=detuning)
    sequence.add_eom_pulse("global", duration=duration, phase=phase)  # type: ignore
    sequence.disable_eom_mode("global")


def dyn_wait(
    sequence: Sequence,
    duration: VariableItem | float,
    *args: Any,
    **kwargs: Any,
) -> None:
    max_amp = sequence.device.channels["rydberg_global"].max_amp or DEFAULT_AMPLITUDE

    duration *= 1000 * 2 * np.pi / max_amp  # type: ignore

    sequence.delay(int(duration), "global")  # type: ignore


def apply_local_shifts(sequence: Sequence, **_: Any) -> None:
    max_abs_detuning = (
        sequence.device.channels["rydberg_global"].max_abs_detuning or DEFAULT_DETUNING
    )
    time_scale = 1000 * 2 * np.pi / max_abs_detuning
    local_pulse_core(sequence, duration=1.0, time_scale=time_scale, detuning=1.0, concurrent=False)


def local_pulse(
    sequence: Sequence,
    duration: VariableItem | float | Duration,
    detuning: VariableItem | float,
    concurrent: bool = False,
    **_: Any,
) -> None:
    max_amp = sequence.device.channels["rydberg_global"].max_amp or DEFAULT_AMPLITUDE
    time_scale = 1000 * 2 * np.pi / max_amp
    local_pulse_core(sequence, duration, time_scale, detuning, concurrent)


def local_pulse_core(
    sequence: Sequence,
    duration: VariableItem | float | Duration,
    time_scale: float,
    detuning: VariableItem | float,
    concurrent: bool = False,
    **kwargs: Any,
) -> None:
    max_abs_detuning = (
        sequence.device.channels["rydberg_global"].max_abs_detuning or DEFAULT_DETUNING
    )

    if duration == Duration.FILL:
        if not concurrent:
            raise SyntaxError("The option `fill` can only be used on the `concurrent` mode")

        duration = sequence.get_duration("global") - sequence.get_duration("dmm_0")
    else:
        duration *= time_scale  # type: ignore

    detuning *= -1 * max_abs_detuning  # type: ignore

    sequence.add_dmm_detuning(
        ConstantWaveform(duration, detuning),
        "dmm_0",
        "no-delay" if concurrent else "min-delay",
    )



from __future__ import annotations

from qadence2_ir.types import Model

from . import register, sequence
from .interface import Interface


def compile_to_backend(model: Model) -> Interface:
    reg = register.from_model(model)
    seq = sequence.from_model(model, reg)
    non_trainable_parameters = {k for k, v in model.inputs.items() if not v.is_trainable}
    return Interface(seq, non_trainable_parameters)



from __future__ import annotations

from abc import ABC, abstractmethod
from enum import Enum, auto
from typing import Any, Callable, Generic, Optional, TypeVar

ArrayType = TypeVar("ArrayType")
SequenceType = TypeVar("SequenceType")
ParameterType = TypeVar("ParameterType")
RunResultType = TypeVar("RunResultType")
SampleResultType = TypeVar("SampleResultType")
ExpectationResultType = TypeVar("ExpectationResultType")


class RunEnum(Enum):
    """
    Enum class to be used whenever an Interface class method need to specify.

    how to execute the expression: through `run`, `sample`, or `expectation`.
    """

    RUN = auto()
    SAMPLE = auto()
    EXPECTATION = auto()


class OnEnum(Enum):
    """
    Enum class to be used whenever an Interface class method (such as `run`) needs to.

    specify where to run the code: on emulator or qpu.
    """

    EMULATOR = auto()
    QPU = auto()


class AbstractInterface(
    ABC,
    Generic[
        ArrayType,
        SequenceType,
        ParameterType,
        RunResultType,
        SampleResultType,
        ExpectationResultType,
    ],
):
    """
    An abstract base class that defines interface essential methods.

    It should be
    inherited by any class that needs to implement a backend, for instance `pyqtorch`
    and `fresnel1` (`pulser` using `qutip` emulator). It is not only used by the package
    itself, but users who want to implement or test new backends should also make use
    of it.
    """

    @property
    @abstractmethod
    def info(self) -> dict[str, Any]:
        """
        Gives any relevant information about the interface data, such as `device`,.

        `register`, etc.

        :return: dictionary with the relevant information.
        """
        pass

    @property
    @abstractmethod
    def sequence(self) -> SequenceType:
        """
        Outputs the backends-native sequence.

        :return: The defined backends-native sequence.
        """
        pass

    @abstractmethod
    def set_parameters(self, params: dict[str, ParameterType]) -> None:
        """
        Sets valid parameters for the backends to use it during simulation/execution step.

        :param params: the fixed parameters to be used by the backends.
        """
        pass

    @abstractmethod
    def run(
        self,
        *,
        values: dict[str, ArrayType] | None = None,
        callback: Optional[Callable] = None,
        **kwargs: Any,
    ) -> RunResultType:
        """
        Gets the results from the expression computation given the parameters (values),.

        callback function (if applicable), and extra arguments.

        :param values: dictionary of user-input parameters
        :param callback: a callback function if necessary to run some extra processing
        :param kwargs: any extra argument that are backends specific can be included in the
            child method.
        :return: any result type according to what is expected by the backends `run` method
        """
        pass

    @abstractmethod
    def sample(
        self,
        *,
        values: dict[str, ArrayType] | None = None,
        shots: int | None = None,
        callback: Optional[Callable] = None,
        **kwargs: Any,
    ) -> SampleResultType:
        """
        Samples the computed result given the expression, the parameters (values), number of.

        shots, callback function (if applicable), and extra arguments.

        :param values: dictionary of user-input parameters
        :param shots: number of shots
        :param callback: a callback function if necessary to run some extra processing
        :param kwargs: any extra argument that are backends specific can be included in the
            child method
        :return: any result type according to what is expected by the backends `sample` method
        """
        pass

    @abstractmethod
    def expectation(
        self,
        *,
        values: dict[str, ArrayType] | None = None,
        observable: Any | None = None,
        callback: Optional[Callable] = None,
        **kwargs: Any,
    ) -> ExpectationResultType:
        """
        Computes the expectation value for observable(s) given the parameters (values),.

        callback function (if applicable), and extra arguments.

        :param values: dictionary of user-input parameters
        :param observable: list of observables
        :param callback: a callback function if necessary to run some extra processing
        :param kwargs: any extra argument that are backends specific can be included in the
            child method
        :return: any result type according to what is expected by the backends `expectation` method
        """
        pass



# type: ignore
# TODO: 1. remove the line above after implementing this file
# TODO: 2. remove these comments

from __future__ import annotations

from qadence2_ir.types import Model

from .interface import Interface


def compile_to_backend(model: Model) -> Interface:
    # TODO: 1. implement this method logic
    # TODO: 2. do not add more arguments
    # TODO: 3. write a docstring if needed
    # TODO: 4. fill the return below with the respective arguments
    # TODO: 5. erase these TODO comments
    return Interface()



from __future__ import annotations

from typing import cast

from qadence2_ir.types import Model

from qadence2_platforms.utils.module_importer import module_loader

from .abstracts import AbstractInterface as Interface


def compile_to_backend(model: Model, backend: str) -> Interface:
    """
    Function that gets a `Model` (Qadence IR) and a backend name, and.

    returns an `Interface` instance from the specific backend with the
    model transformed into backend appropriate data.

    :param model: (Model) qadence IR
    :param backend: (str) the backend to be used to execute the Model
    :return: (Interface) interface instance of the chosen backend
    """

    plat = module_loader(backend)
    return cast(Interface, plat.compile_to_backend(model))



from __future__ import annotations

import os
import sys
import traceback
from pathlib import Path
from shutil import copyfile

from qadence2_platforms import CUSTOM_BACKEND_FOLDER_NAME, TEMPLATES_FOLDER_NAME
from qadence2_platforms.utils.module_importer import resolve_module_path

try:
    from tkinter import filedialog as fd

except ImportError:

    def user_input() -> str:
        return input("Paste the directory: ")

else:

    def user_input() -> str:
        return fd.askdirectory(
            initialdir=Path(__file__).parent,
            title="Select a directory",
            mustexist=True,
        )


class BackendTemplate:
    """
    Class to create new custom backend folder with file templates.

    It will
    follow the same structure as the built-in backends, namely fresnel1 and
    pyqtorch, having a folder with the custom backend name with the essential
    files inside already built with the functions, classes and their methods.

    Ex:

    selected_root_dir/
    └── custom_backends/
        ├── custom_backend1/
        │   ├── __init__.py
        │   ├── compiler.py
        │   └── interface.py

    It is intended to help the user when creating their custom backend with
    pre-filled files so the core structure of the backend instances are still
    present, while giving freedom for them to implement whatever else needed.
    """

    def __init__(self) -> None:
        self._pwd: Path = Path()
        self._root_backends_name: str = CUSTOM_BACKEND_FOLDER_NAME
        self._backend_name: str = ""
        self._backend_path: Path = Path()
        self._template_path: Path = self._get_template_path()
        self._template_files_list: list[str] = os.listdir(self._template_path)

    @property
    def user_backend_path(self) -> Path:
        return self._backend_path

    @property
    def platforms_backend_path(self) -> Path:
        return Path()

    @property
    def template_files_list(self) -> list[str]:
        return self._template_files_list

    def _new_file_path(self, name: str) -> Path:
        return self._backend_path / name

    def _get_template_path(self) -> Path:
        return Path(__file__).parent / TEMPLATES_FOLDER_NAME

    def create_folder(self, backend_name: str, current_path: str | Path) -> bool:
        """
        Creates the main folder for the custom backend in a selected path.

        :param backend_name: (str) backend name
        :param current_path: (str | Path) current path to place the custom backend
        :return: (bool) returns true if the folder was already existing
        """

        self._backend_name = backend_name
        self._pwd = Path(current_path)
        self._backend_path = Path(self._pwd, self._root_backends_name, self._backend_name)
        already_exists = os.path.exists(self._backend_path)
        self._backend_path.mkdir(parents=True, exist_ok=True)
        return already_exists

    def create_files(self) -> None:
        for file in self.template_files_list:
            self._new_file_path(file).touch()
            copyfile(self._template_path / file, self._backend_path / file)

    def create_template(
        self,
        backend_name: str,
        gui: bool = True,
        use_this_dir: str | Path | None = None,
    ) -> bool:
        """
        Creates the template, with the main custom backend folder and its content files.

        :param backend_name: (str) backend name
        :param gui: (bool) whether to use a GUI option to choose where to create the template;
             tkinter must be installed.
             On Mac: `brew install python-tk@python3.10`. In case you use a different python
             version, replace `3.10` by it. On Linux: `apt-get install python-tk`.
        :param use_this_dir: (str | Path | None) directory to create the custom backend
        :return: (bool) returns true if the template was successfully created
        """

        print(
            "\nCreating a backend template.\n\n"
            "You need to select a directory where all the custom backends will be located.\n"
            " The organization will be as follows:\n\n"
            "   selected_root_dir/\n"
            "   └── custom_backends/\n"
            "       ├── custom_backend1/\n"
            "       │   ├── __init__.py\n"
            "       │   ├── compiler.py\n"
            "       │   └── interface.py\n"
            "       ├── custom_backend2/\n"
            "       ...\n"
            ""
        )

        selected_dir: str | Path

        if gui and use_this_dir is None and "tkinter" in sys.modules:
            selected_dir = user_input()
        else:
            if use_this_dir is None:
                selected_dir = user_input()
            else:
                selected_dir = use_this_dir

        if selected_dir:
            try:
                already_exists = self.create_folder(backend_name, current_path=selected_dir)
                self.create_files()
            except Exception:
                traceback.print_exc()
                return False
            else:
                action = "replaced" if already_exists else "created"
                result = resolve_module_path(Path(selected_dir, CUSTOM_BACKEND_FOLDER_NAME))
                if result:
                    print(
                        f"Backend template at {self._backend_path} has been {action} with success!"
                    )
                return result
        else:
            print("You must select a directory to create the template. Creation suspended.")
            return False



from __future__ import annotations

from typing import cast

import pyqtorch as pyq
from pyqtorch.hamiltonians import Observable
from pyqtorch.primitives import Primitive

from qadence2_platforms.backends.pyqtorch.utils import InputType, Support


def _get_op(op: InputType) -> Primitive | None:
    if op.is_symbol is True:
        return getattr(pyq, op.head.upper(), None)

    if op.is_quantum_operator is True:
        op_args_item: InputType = op.args[0].args
        return getattr(pyq, op_args_item[0].upper(), None)

    op_args: str = cast(str, op.args[0])
    return getattr(pyq, op_args.upper(), None)


def _get_native_op(op: InputType) -> Primitive:
    native_op = _get_op(op)
    if native_op is not None:
        support: Support = cast(Support, op.args[1])
        return native_op(*support.target, support.control)

    raise ValueError(f"Observable {op} not found")


def load_observables(observable: list[InputType] | InputType) -> Observable:
    if isinstance(observable, list):
        pyq_observables: list[Primitive] | list = []
        for op in observable:
            pyq_observables.append(_get_native_op(op))
        return Observable(*pyq_observables)

    return Observable(_get_native_op(observable))



from __future__ import annotations


class RegisterInterface:
    def __init__(self, n_qubits: int, init_state: str | None = None) -> None:
        self.n_qubits = n_qubits
        self.init_state = init_state
